;
;
; SNEeSe SNES memory mapper v2.0 - NASM rewrite
;
; Notes: LoROM has max ROM address space of 4MBytes.
;
; LoROM version
;  00-3F:0000-1FFF = Low RAM 8k (7E:0000-1FFF mirrored)
;  00-3F:2000-5FFF = Hardware ports (PPU, FX, DSP)
;  00-3F:6000-7FFF = Unmapped
;  00-3F:8000-FFFF = ROM 32k (0-16Mbit)
;  40-6F:8000-FFFF = ROM 32k (16-28Mbit)
;  70-7D:0000-FFFF = SRAM (up to 64k) (continuously mirrored: see below)
;  7E-7F:0000-FFFF = WRAM 128k
;  80-EF:0000-FFFF = 00-6F mirrored
;  F0-FF:8000-FFFF = ROM 32k (28-32Mbit)
;
;  SRAM has 256k (kbytes, not kbits) of address space. This is mapped
;   with 'continuous mirroring', as such:
;    2k (16kbit): 70:0000-07FF == 70:0800-0FFF == 73:F800-FFFF
;    8k (64kbit): 70:0000-1FFF == 70:2000-3FFF == 73:E000-FFFF
;   etc.
;  Since I know of no ROMs that require >64k of SRAM, only 64k is
;   supported (however, 256k mirroring is performed).
;  Note: I HAVE discovered one ROM requiring 128k of SRAM (Dezaemon),
;   will implement this later...
;

%define SNEeSe_memlo_asm

%include "misc.ni"
%include "memmap.ni"
%include "cpumem.ni"
%include "PPU.ni"

section .text
EXPORT_C memlo_text_start
section .data
EXPORT_C memlo_data_start
section .bss
EXPORT_C memlo_bss_start

extern _SRAM_Mask

section .data
; _LoROM_Read is _LoROM_Blank - see MemBlank.asm

ALIGND
EXPORT_C LoROM_Write

; Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_WRITE_00_3F

; Banks 40-6F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,48,UNSUPPORTED_WRITE

; Banks 70-7D is SRAM (mirrored)
    DUPLICATE dd,14,_SRAM_WRITE

; Bank  7E-7f is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_WRITE

; Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_WRITE_80_BF

; Banks C0-FF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,UNSUPPORTED_WRITE

section .text
; ebx - contains the actual address, al is where the info should be stored, edi is free
; NB ebx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

ALIGNC
EXPORT_C SRAM_READ
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov al,[_SRAM+edi]
    ret

ALIGNC
EXPORT_C SRAM_WRITE
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov [_SRAM+edi],al
    ret

ALIGNC
EXPORT_C SRAM_WRITE_2k
    mov edi,2048 - 1
    and edi,ebx
    mov [_SRAM+edi],al
    mov [_SRAM+edi+2048],al
    mov [_SRAM+edi+4096],al
    mov [_SRAM+edi+6144],al
    ret

ALIGNC
EXPORT_C SRAM_WRITE_4k
    mov edi,4096 - 1
    and edi,ebx
    mov [_SRAM+edi],al
    mov [_SRAM+edi+4096],al
    ret

; ebx - contains the actual address, al is where the info should be stored, edi is free
; NB ebx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

ALIGNC
EXPORT_C READ_00_3F_2
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k_2
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

ALIGNC
EXPORT_C READ_40_6F_2
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k_2
ALIGNC
.not_rom:
    NON_MAPPED_READ

ALIGNC
EXPORT_C READ_80_BF_2
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k_2
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

ALIGNC
EXPORT_C READ_C0_FF_2
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k_2
ALIGNC
.not_rom:
    NON_MAPPED_READ

section .text
ALIGNC
section .data
ALIGND
section .bss
ALIGNB
