;
; Sprite render functions
;
; In all the sprite renderers...
;  eax = (internal)
;  ebx = (internal)
;  cl  = sprite counter for priority (internal)
;  ch  = sprite counter for set      (internal)
;  dl  = sprite priority identifier  (input)
;  dh  = (internal)
;  esi = current sprite              (internal)
;  edi = (internal)

;#define Profile_Recache_OAM

%include "misc.ni"
%include "screen.ni"
%include "PPU.ni"

section .text
EXPORT_C sprites_text_start
section .data
EXPORT_C sprites_data_start
section .bss
EXPORT_C sprites_bss_start

extern Ready_Line_Render,ScreenDestPtr
extern _TileCache4
extern Tile_Offset_Table_64_8
extern Sprite_Tile_Plot_8_4,Sprite_Tile_Plot_16_4
extern Sprite_Tile_Plot_32_4,Sprite_Tile_Plot_64_4

section .data
ALIGND
; Sprite offset tables moved to ScreenL.S
Sprite_Mode_Table:
dd RENDER_SPRITE_8x8_16x16      ; 8x8 or 16x16
dd RENDER_SPRITE_8x8_32x32      ; 8x8 or 32x32
dd RENDER_SPRITE_8x8_64x64      ; 8x8 or 64x64
dd RENDER_SPRITE_16x16_32x32    ; 16x16 or 32x32
dd RENDER_SPRITE_16x16_64x64    ; 16x16 or 64x64
dd RENDER_SPRITE_32x32_64x64    ; 32x32 or 64x64
dd SPRITE_NOT_SUPPORTED         ; Not supported in SNES hardware?
dd SPRITE_NOT_SUPPORTED         ; Not supported in SNES hardware?

Sprite_Size_Table:
db  8, 16, -7,-15
db  8, 32, -7,-31
db  8, 64, -7,-63
db 16, 32,-15,-31
db 16, 64,-15,-63
db 32, 64,-31,-63
db  0,  0,  0,  0
db  0,  0,  0,  0

section .bss
ALIGNB
EXPORT OAM_X      ,skipl 128    ; OAM sprite X positions
EXPORT OAM_Tile   ,skipl 128    ; OAM sprite tile addresses
EXPORT OAM_Size   ,skipb 128    ; OAM sprite size flags
EXPORT OAM_Y      ,skipb 128    ; OAM sprite Y positions
EXPORT OAM_Palette,skipb 128    ; OAM sprite palettes
EXPORT OAM_Flip   ,skipb 128    ; OAM sprite X/Y flip
OAM_Priority:skipb 128      ; OAM sprite priorities
EXPORT_C OAM,skipb 512+32   ; Buffer for OAM
EXPORT SpriteCount,skipl
EXPORT Sprite_Mode,skipl    ; Renderer for current sprite sizes
EXPORT_C HiSprite ,skipl
EXPORT SPRLatch   ,skipb    ; Sprite Priority Rotation latch flag
EXPORT_C OBSEL    ,skipb    ; sssnnxbb  sss=sprite size,nn=upper 4k address,bb=offset
EXPORT OAMOdd     ,skipb
EXPORT OAMReadOdd ,skipb
EXPORT_C HiSpriteCnt1,skipl ; First set size and bit offset
EXPORT_C HiSpriteCnt2,skipl ; Second set size and bit offset
EXPORT_C OBBASE,skipl   ; VRAM location of sprite tiles 00-FF
EXPORT_C OBNAME,skipl   ; VRAM location of sprite tiles 100-1FF
EXPORT_C OAMAddress,skipl
EXPORT_C OAMAddress_VBL,skipl   ; Restore this at VBL
EXPORT_C HiSpriteAddr,skipl     ; OAM address of sprite in 512b table
EXPORT_C HiSpriteBits,skipl     ; OAM address of sprite in 32b table
EXPORT NextSprite   ,skipl
EXPORT Priority_Used,skipl
ALIGNB
EXPORT SpriteZ,skipb 64+256+64
ALIGNB
EXPORT Rendered_Sprites,skipb 64+256+64
ALIGNB
EXPORT SpriteList,skipb 129
EXPORT Redo_OAM,skipb

section .text
ALIGNC
; This covers special cases or not implemented features!
SPRITE_NOT_SUPPORTED:
 mov eax,[Priority_Used]
 test eax,eax
 jz .no_clear
  
 mov edi,Rendered_Sprites+64
 xor eax,eax
 mov ebp,256/32
 mov [Priority_Used],eax
 jmp .clear_loop

ALIGNC
.clear_loop:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop

 mov edi,SpriteZ+64
 mov ebp,256/32
 dec eax
 jmp .clear_loop_z

ALIGNC
.clear_loop_z:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop_z

.no_clear:
 ret

%macro PlotSprite 1
; Don't draw unless sprite is on this scanline
 mov al,[Current_Line_Render]
 dec eax
 sub al,[OAM_Y+esi] ; Get Y position
 cmp al,%1
 jae %%check_count
 mov bl,[OAM_Priority+esi]
 mov dl,[OAM_Flip+esi]
 and ebx,byte 3
 and eax,byte 0x3F
 mov cl,bl
 mov dh,[OAM_Palette+esi]
 mov byte [Priority_Used+ebx],-1
 mov ebp,[OAM_X+esi*4]  ; X position
 mov edi,Rendered_Sprites+64
;mov edi,[ScreenDestPtr]
 mov ebx,SpriteZ+64
 add edi,ebp            ; X position in render-buffer
 add ebx,ebp            ; X position in Z-buffer
 mov ebp,edx
 mov esi,[OAM_Tile+esi*4]
;mov [ColourBase],dh

 ; We have offset into sprite offset table
 mov eax,[Tile_Offset_Table_64_8+eax*4]
 test ebp,8
 jz %%no_yflip

 xor eax,byte -1
%if %1-8  ; Not 8x8
%if %1-16 ; Not 16x16
%if %1-32 ; Not 32x32
 add eax,1+(7*16*64+7*8)
%else
 add eax,1+(3*16*64+7*8)
%endif
%else
 add eax,1+(16*64+7*8)
%endif
%else
 add eax,byte 1+(7*8)
%endif

%%no_yflip:
 add esi,eax
 and ebp,byte 4 ; ebp = X

%if %1-8  ; Not 8-bit
%if %1-16 ; Not 16-bit
%if %1-32 ; Not 32-bit
  call dword [Sprite_Tile_Plot_64_4+ebp]
%else
  call dword [Sprite_Tile_Plot_32_4+ebp]
%endif
%else
  call dword [Sprite_Tile_Plot_16_4+ebp]
%endif
%else
  call dword [Sprite_Tile_Plot_8_4+ebp]
%endif

%%check_count:
 dec ch
 jnz near .next_sprite  ; Loop for all sprites in set
 ret
%endmacro

%macro RenderSprite 3
ALIGNC
EXPORT RENDER_SPRITE_%3
 mov ch,al

 mov eax,[Priority_Used]
 test eax,eax
 jz near .next_sprite
  
 mov edi,Rendered_Sprites+64
 xor eax,eax
 mov ebp,256/32
 mov [Priority_Used],eax
 jmp .clear_loop

ALIGNC
.clear_loop:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop

 mov edi,SpriteZ+64
 mov ebp,256/32
 dec eax
 jmp .clear_loop_z

ALIGNC
.clear_loop_z:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop_z
 jmp .next_sprite

ALIGNC
.next_sprite:
 mov esi,[NextSprite]
 mov esi,[esi]
 and esi,0xFF
 inc dword [NextSprite]
 mov al,[OAM_Size+esi]
 test al,al
 jnz near .do_large
 PlotSprite %1

ALIGNC
.do_large:
 PlotSprite %2

%endmacro

RenderSprite 8,16,8x8_16x16
RenderSprite 8,32,8x8_32x32
RenderSprite 8,64,8x8_64x64
RenderSprite 16,32,16x16_32x32
RenderSprite 16,64,16x16_64x64
RenderSprite 32,64,32x32_64x64

%macro Plot_Sprites_2 1
  mov al,[%1+ebx]
  mov dl,[%1+4+ebx]
  cmp cl,al
  jne %%no_plot_0
  mov al,[%1+edi]
  test al,al
  jnz %%no_plot_0
  mov al,[%1+esi]
  mov [%1+edi],al
%%no_plot_0:
  cmp cl,dl
  jne %%no_plot_4
  mov dl,[%1+4+edi]
  test dl,dl
  jnz %%no_plot_4
  mov dl,[%1+4+esi]
  mov [%1+4+edi],dl
%%no_plot_4:
%endmacro

ALIGNC
EXPORT Plot_Sprites
 mov ebx,SpriteZ+64
 mov esi,Rendered_Sprites+64
 mov edi,[ScreenDestPtr]
 mov ebp,256/8

.next_8:
 Plot_Sprites_2 0
 Plot_Sprites_2 1
 Plot_Sprites_2 2
 Plot_Sprites_2 3

 add ebx,byte 8
 add esi,byte 8
 add edi,byte 8
 dec ebp
 jnz near .next_8
 ret

; In the precalculator...
;  eax = (internal)
;  edx = (internal)
;  ebx = current sprite          (internal)
;  cl  = sprite flags bit count  (internal)
;  ebp = (internal)
;  esi = sprite tables           (internal)
;  edi = sprite flags            (internal)
;
ALIGNC
EXPORT_C Recache_OAM
EXPORT Recache_OAM
; Need to rewrite this to calculate in order of render, and
;  to generate sprite lists for each priority!
;  (as this is only called when sprites are enabled, and OBSEL, OAM, or
;   highest priority sprite has been changed, and each layer would have
;   to decipher list on its own, added overhead should be minimal when
;   called and easily made up for when not called
%ifdef Profile_Recache_OAM
 inc dword [_Calls_Recache_OAM]
%endif
 xor edx,edx
 mov ebx,[_HiSprite]
 mov [SpriteList],dl
 mov ecx,[_HiSpriteCnt1]    ; Size of first set and bit offset
 mov eax,[_HiSpriteCnt2]
 mov esi,[_HiSpriteAddr]
 test ah,ah
 mov edi,[_HiSpriteBits]
 jz .is_zero
 push dword 0
.is_zero:
 push eax
 jmp .next_sprite

; Tile attribute word: YXPP CCCT TTTT TTTT
;  Where:
;   Y, X are vertical/horizontal flip
;   P is priority
;   C is color palette selector
;   T is tile number
;
ALIGNC
.next_sprite:
 mov dh,[edi]
 shl dh,cl
 jnc .do_large
 add dh,dh
 sbb eax,eax
 mov al,[esi]   ; eax = X
 mov byte [OAM_Size+ebx],0xFF
 cmp eax,byte -15
global sprlim_large
sprlim_large equ $-1
 jge .size_done
 jmp .off_screen

ALIGNC
.do_large:
 add dh,dh
 sbb eax,eax
 mov al,[esi]  ; eax = X
 mov byte [OAM_Size+ebx],0
 cmp eax,byte -7
global sprlim_small
sprlim_small equ $-1
 jl .off_screen
.size_done:
 mov dl,[3+esi]
 mov [OAM_X+ebx*4],eax
 shr edx,4
 and edx,byte 3
 mov [OAM_Priority+ebx],dl
 xor eax,eax
 mov al,[SpriteList]
 inc eax
 mov [SpriteList],al
 mov [SpriteList+eax],bl
 mov al,[esi+1]
 mov dl,[esi+2]
 mov [OAM_Y+ebx],al
 mov dh,[esi+3]

 mov al,dh
 and dh,1

 shl edx,6  ; tile * 64
 mov ah,al
 shl al,3
 shr ah,4
 or al,0x8F ; ColourBase = 1CCC 1111
 mov [OAM_Tile+ebx*4],edx
 and ah,0x0C    ; 0000 yx00
 mov [OAM_Palette+ebx],al
 mov [OAM_Flip+ebx],ah
 jmp .off_screen

ALIGNC
.off_screen:
 sub cl,2
 jns .bits_left
 and cl,7
 inc edi    ; Goto next sprite byte

.bits_left:
 add esi,byte 4
 inc bl
 dec ch
 jnz near .next_sprite
 mov ch,[1+esp]
 sub ebx,0x80
 sub esi,0x200
 sub edi,byte 0x20
 add esp,byte 4
 test ch,ch
 jnz near .next_sprite

 mov byte [Redo_OAM],0

 mov eax,[Priority_Used]
 test eax,eax
 jz .no_clear
 mov al,[SpriteList]
 test al,al
 jnz .no_clear
 xor eax,eax
  
 mov edi,Rendered_Sprites+64
 xor eax,eax
 mov ebp,256/32
 mov [Priority_Used],eax
 jmp .clear_loop

ALIGNC
.clear_loop:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop

 mov edi,SpriteZ+64
 mov ebp,256/32
 dec eax
 jmp .clear_loop_z

ALIGNC
.clear_loop_z:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop_z

.no_clear:
 ret

ALIGNC
EXPORT_C Reset_Sprites
 pusha
 ; Set eax to 0, as we're setting most everything to 0...
 xor eax,eax

 ; Reset sprite renderer vars
 mov dword [_HiSprite],0
 mov dword [_HiSpriteAddr],_OAM+0x000
 mov dword [_HiSpriteBits],_OAM+0x200
 mov dword [_HiSpriteCnt1],0x8007
 mov dword [_HiSpriteCnt2],0x0007
 mov dword [Sprite_Mode],RENDER_SPRITE_8x8_16x16
 mov byte [Redo_OAM],-1
 mov byte [sprlim_small],-7
 mov byte [sprlim_large],-15
 mov [_OBBASE],eax
 mov [_OBNAME],eax

 ; Reset sprite port vars
 mov [_OAMAddress],eax
 mov [_OAMAddress_VBL],eax
 mov [OAMOdd],al
 mov [OAMReadOdd],al
 mov [SPRLatch],al
 mov [_OBSEL],al

;xor eax,eax
 mov edi,Rendered_Sprites+64
 mov ebp,256/32
 mov [Priority_Used],eax
 jmp .clear_loop

ALIGNC
.clear_loop:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop

 mov edi,SpriteZ+64
 mov ebp,256/32
 dec eax
 jmp .clear_loop_z

ALIGNC
.clear_loop_z:
 lea esi,[edi+16]
 mov [edi],eax
 mov [edi+4],eax
 mov [edi+8],eax
 mov [edi+12],eax
 mov [esi],eax
 lea edi,[esi+16]
 mov [esi+4],eax
 mov [esi+8],eax
 mov [esi+12],eax
 dec ebp
 jnz .clear_loop_z

 popa
 ret

ALIGNC
EXPORT SNES_R2138 ; OAMDATAREAD
 push ebx
 xor ebx,ebx
 mov bl,[OAMReadOdd]
 mov edi,[_OAMAddress]
 mov al,[_OAM+ebx+edi*2]
 xor bl,1
 mov [OAMReadOdd],bl
 jnz .no_increment
 inc edi
 cmp edi,0x110  ; Ensure address wraps correctly!
 je .wrap_address
 mov [_OAMAddress],edi
.no_increment:
 pop ebx
 ret
ALIGNC
.wrap_address:
 mov dword [_OAMAddress],0
 pop ebx
 ret

section .bss
EXPORT_C OBSEL9_WRITE,skipl
extern _PBOffset
extern __65816_PB_Shifted

section .text
ALIGNC
EXPORT SNES_W2101 ; OBSEL
 cmp [_OBSEL],al
 je near .no_change

 cmp al,9
 jne .no_fault
 mov edi,esi
 sub edi,[_PBOffset]
 or edi,[__65816_PB_Shifted]
 mov [_OBSEL9_WRITE],edi
.no_fault:

 SyncRender     ;*
 push ebx
 mov [_OBSEL],al    ; Get our copy of this
 mov ebx,eax
 shr eax,5
 and eax,byte 7
 push edx
 mov dl,[Sprite_Size_Table+2+eax*4]
 mov dh,[Sprite_Size_Table+3+eax*4]
 mov [sprlim_small],dl
 mov [sprlim_large],dh
 pop edx
 mov eax,[Sprite_Mode_Table+eax*4]
 mov [Sprite_Mode],eax
 mov eax,ebx
 mov edi,eax
 and ebx,byte 3<<3  ; Name address 0000 0000 000n n000
 and edi,byte 3     ; Base address 0000 0000 0000 0xbb
 shl ebx,10         ; Name is either 0x0000,0x1000,0x2000,0x3000
 shl edi,14         ; Base is either 0x0000,0x2000,0x4000,0x6000
 add ebx,edi
 and edi,0xFFFF
 and ebx,0xFFFF
 add edi,edi        ; Convert to offsets into tile cache
 add ebx,ebx
;add edi,[_TileCache4]
;add ebx,[_TileCache4]
 mov [_OBBASE],edi
 mov [_OBNAME],ebx
 pop ebx
 mov byte [Redo_OAM],-1
.no_change:
 ret

ALIGNC
EXPORT SNES_W2102 ; OAMADDL
 SyncRender     ;*
 push ebx
 xor ebx,ebx
 mov [OAMOdd],bh
 mov bl,al
 mov [OAMReadOdd],bh
 mov [_OAMAddress],ebx
 mov [_OAMAddress_VBL],ebx
 cmp byte [SPRLatch],0
 jz .no_latch_change
;add ebx,ebx    ;*
;sub ebx,byte 4
.update_latch:
 and ebx,byte 0x7F  ;*
 cmp [_HiSprite],ebx
 je .no_latch_change
 mov [_HiSprite],ebx
 shl ebx,2
 mov [_HiSpriteAddr],ebx
 shr ebx,4
 add ebx,0x200
 mov [_HiSpriteBits],ebx
 mov ebx,[_HiSprite]
 dec ebx    ;
 mov bh,128
 inc bl
 sub bh,bl
 mov [_HiSpriteCnt2+1],bl   ;
 mov [_HiSpriteCnt1+1],bh   ;
 mov bh,7
 add bl,bl
 sub bh,bl
 and bh,7
 mov [_HiSpriteCnt1],bh ;

 mov ebx,_OAM
 add [_HiSpriteAddr],ebx
 add [_HiSpriteBits],ebx
.no_latch_change:
 pop ebx
 ret

ALIGNC
EXPORT SNES_W2103 ; OAMADDH
 SyncRender     ;*
 push ebx
 xor ebx,ebx
 test al,al     ; Is priority rotation bit set?
 mov [OAMOdd],bl
 mov bh,1
 mov [OAMReadOdd],bl
 jns .no_latch_change
 mov bl,-1
.no_latch_change:
 and bh,al      ; Only want MSB of address
 mov [SPRLatch],bl
 test bl,bl
 mov bl,[_OAMAddress]
 mov [_OAMAddress_VBL+1],bh
 mov [_OAMAddress+1],bh
 jnz near SNES_W2102.update_latch
 pop ebx
 ret

ALIGNC
EXPORT SNES_W2104 ; OAMDATA
 push ebx
 xor ebx,ebx
 mov bl,[OAMOdd]
 mov edi,[_OAMAddress]
 cmp edi,0x110
 jae .ignore_write
 lea edi,[ebx+edi*2]
 cmp [_OAM+edi],al
 je .no_change
 push edi
 SyncRender     ;*
 pop edi
 mov byte [Redo_OAM],-1
 mov [_OAM+edi],al
.no_change:
 xor bl,1
 mov [OAMOdd],bl
 jnz .no_increment
 mov edi,[_OAMAddress]
 inc edi
 cmp edi,0x0110     ; Ensure address wraps correctly!
 je .wrap_address
 mov [_OAMAddress],edi
.no_increment:
.ignore_write:
 pop ebx
 ret
ALIGNC
.wrap_address:
 mov dword [_OAMAddress],0
 pop ebx
 ret

section .text
ALIGNC
section .data
ALIGND
section .bss
ALIGNB
